             .title        "MMC4 Reader Plugin"


;10/21/06
;Written by BootGod
;Version 1.0

;Note: FJROM doesn't have WRAM enabled on powerup and I don't know how to do it!
;A way to get around this is, while testing game just before dumping, play into the
;game enough (get thru all menus) and let the game enable it for you. It will stay
;enabled for when you dump as long as you dont turn the power off. FKROM does not
;have this issue.
             
;vectors for CopyNES BIOS calls

send_byte:   .equ 0200h		;sends byte to client
baton:       .equ 0203h		;indicates xfer progress on LCD if installed
chk_wram:    .equ 0209h		;checks for presence of WRAM
wr_ppu:      .equ 020ch		;sets address for VRAM access
init_crc:    .equ 0212h
do_crc:      .equ 0215h
finish_crc:  .equ 0218h

crc0:        .equ 0080h
crc1:        .equ 0081h
crc2:        .equ 0082h
crc3:        .equ 0083h

;variables

address:     .equ 00e0h
address_lo:  .equ 00e0h
address_hi:  .equ 00e1h
banksleft:   .equ 00e2h
curbank:     .equ 00e3h
numpages_lo  .equ 00e4h
numpages_hi  .equ 00e5h
bank0crc:    .equ 00e8h
temp_crc:    .equ 00ech
tx:          .equ 00e6h


;plugin header defines
p_ver: = 1	;plugin version
p_read: = 1	;plugin is for downloading (dumping)
p_write: = 2	;plugin is for uploading (RAM carts / WRAM etc)
p_other: = 3	;plugin is for some other purpose (i.e. BankWatch)
p_var1: = 4	;plugin uses user variable #1
p_var2: = 8	;plugin uses user variable #2
p_var3: = 16	;plugin uses user variable #3
p_var4: = 32	;plugin uses user variable #4

;plugin header data

             .org 0380h
             
             ;plugin description
             .db "Nintendo FxROM / MMC4 carts"
             .fill 03e0h-*,00h

             ;plugin creator(s)
             .db "BootGod"
             .fill 03f8h-*,00h

             ;PRG/CHR/WRAM min/max sizes supported
             .db 128 >> 4,256 >> 4
             .db 64 >> 3,128 >> 3
             .db 8,0

             ;plugin version and flags
             .db p_ver
             .db p_read

;plugin code start here

;mirroring detection
;Horizontal	(x = 1)
;Vertical	(x = 0)
;Four-Screen	(x = 2)
;MapperCtrl	(x = 4)

send_mirr:   lda #04h		;MMC4 mirroring is controlled by mapper
             jsr send_byte

;get crc of PRG bank 0 (16K)

             lda #0
             sta 0a000h		;set to PRG bank 0
             jsr crc_pbank

             ldx #3
ld_pcrc:     lda temp_crc,x	;save for comparision
             sta bank0crc,x
             dex
             bpl ld_pcrc

;start out crc check at bank 8	;128K in
             lda #8
             sta curbank
             lda #2
             sta numpages_hi

next_pcrc:   lda curbank
             sta 0a000h		;set PRG bank num
             jsr crc_pbank
             
;compare crc's of bank 0 and curbank
             ldx #3
chk_crc:     lda bank0crc,x
             cmp temp_crc,x
             bne crc_fail
             dex
             bpl chk_crc

             jmp got_prg

crc_fail:    lda curbank	;check to see if we've hit the max bank num
             cmp #16
             beq got_prg

             asl numpages_hi	;size in pages to send to pc
             asl curbank
             jmp next_pcrc

;Set up for PRG transfer

got_prg:     lda #0
             jsr send_byte	;send numpages_lo (0)
             lda numpages_hi
             jsr send_byte	;send numpages_hi

             lda numpages_hi	;# of 16K banks
             asl a
             asl a
             sta banksleft
             lda #0
             sta curbank

             lda #1		;send type is PRG
             jsr send_byte

next_prg:    lda curbank
             sta 0a000h		;write bank num to MMC4 PRG select
             inc curbank
             lda #0		;set read address to $8000
             sta address_lo
             lda #080h
             sta address_hi

             ldx #64		;set counters for 16K bank xfer
             ldy #0

send_prg:    lda (address),y
             jsr send_byte
             iny
             bne send_prg
             inc address_hi
             dex
             bne send_prg
             dec banksleft
             bne next_prg	;16K has been sent, load next bank
             
;determine CHR ROM size

             lda #0
             sta 0b000h		;set to CHR bank 0
             sta 0c000h
             jsr crc_cbank

             ldx #3
ld_ccrc:     lda temp_crc,x	;save for comparision
             sta bank0crc,x
             dex
             bpl ld_ccrc
             
;start out crc check at bank 16 (64K in)
             lda #16
             sta curbank
             lda #1
             sta numpages_hi
             lda #0
             sta numpages_lo

next_ccrc:   lda curbank
             sta 0b000h		;set CHR bank num
             sta 0c000h
             jsr crc_cbank
             
;compare crc's of bank 0 and curbank
             ldx #3
chk_ccrc:    lda bank0crc,x
             cmp temp_crc,x
             bne crc_cfail
             dex
             bpl chk_ccrc

             jmp got_chr	;it's 64K

crc_cfail:   asl numpages_hi	;it's 128K
                         
got_chr:     lda numpages_lo
             jsr send_byte	;send numpages_lo (0)
             lda numpages_hi
             jsr send_byte	;send numpages_hi

             lda numpages_hi	;convert numpages to # of 4K banks (<< 4)
             asl a
             asl a
             asl a
             asl a
             sta banksleft
             lda #0
             sta curbank

             lda #2		;send type is CHR
             jsr send_byte

next_chr:    lda curbank
             sta 0b000h
             sta 0c000h		;write bank num to MMC4 CHR select (write to both so latch0 value doesn't matter)

             lda #00h		;init ppu read
             jsr wr_ppu
             lda 2007h

             ldx #16		;set counters for 4K bank xfer
             ldy #0

send_chr:    lda 2007h
             jsr send_byte
             iny
             bne send_chr
             dex
             bne send_chr
             inc curbank
             dec banksleft
             bne next_chr	;4K has been sent, load next bank

do_wram:     jsr chk_wram
             bne done
             lda #020h
             tax
             jsr send_byte
             lda #0
             jsr send_byte
             lda #3
             jsr send_byte
             lda #0
             sta address_lo
             tay
             lda #060h
             sta address_hi
            
send_wram:   lda (address),y
             jsr send_byte
             iny
             bne send_wram
             inc address_hi
             dex
             bne send_wram

;Signal client no more is coming

done:        lda #0
             jsr send_byte
             lda #0
             jsr send_byte
             lda #0
             jsr send_byte

             rts


;CRC routine for 16K PRG bank @ $8000, stores crc in temp_crc

crc_pbank:   lda #0
             sta address_lo
             lda #080h
             sta address_hi	;set address to $8000

             jsr init_crc
             ldy #0
             ldx #64		;init crc for 16K of data (64 pages = 16K)

crc_ploop:   lda (address),y	;crc generation loop
             jsr do_crc
             iny
             bne crc_ploop
             inc address_hi
             dex
             bpl crc_ploop

             jsr finish_crc	;crc done

             ldx #3
crc_psave:   lda crc0,x		;save crc in temp_crc
             sta temp_crc,x
             dex
             bpl crc_psave

             rts

;CRC routine for 4K CHR bank @ $0000, stores crc in temp_crc

crc_cbank:   lda #000h
             jsr wr_ppu
             lda 2007h		;set PPU address to $0000

             jsr init_crc
             ldx #16
             ldy #0		;init crc for 1K of data (16 pages = 4K)

crc_cloop:   lda 2007h		;crc generation loop
             stx tx
             jsr do_crc
             ldx tx
             iny
             bne crc_cloop
             dex
             bne crc_cloop
             
             jsr finish_crc	;crc done

             ldx #3
crc_csave:   lda crc0,x		;save crc in temp_crc
             sta temp_crc,x
             dex
             bpl crc_csave

             rts


             .fill 0800h-*,0ffh	;fill rest to get 1K of data

             .end
